#ifndef _ARCH_USERCOPY_H
#define _ARCH_USERCOPY_H

#include <kern/types.h>
#include <kern/libc.h>

namespace Arch {
	template <typename T>
	static inline void copyin(T *uptr, T *kdata, int count)
	{
		// OPT: use alignof for movsw and movsl versions
		ulong bytes = sizeof(*kdata) * count;
		int cause;
		
		if ((ulong)uptr + bytes >= PHYSMEM_START ||
		    (ulong)uptr + bytes < (ulong)uptr) {
			cause = 1;
			goto bad;
		}
		
		asm("1: rep; movsb\n"
		    "2:\n"
		    ".section .extable,\"a\"\n"
		    ".align 4\n"
		    ".long 1b\n"
		    ".long 2b\n"
		    ".previous\n" : "+c" (bytes), "+D" (kdata), "+S" (uptr), "=d" (cause));
	
		if (bytes != 0) {
		bad:
			throw_idl(MemoryFault, reinterpret_cast<ulong>(uptr),
			          0, NULL, NULL, cause);
		}
	}

	template <typename T>
	static inline void copyin(T *uptr, T &kdata)
	{
		// OPT: special versions for common small sizes
		copyin(uptr, &kdata, 1);
	}
	
	template <typename T>
	static inline T copyin(T *uptr)
	{
		T kdata;
		copyin(uptr, kdata);
		return kdata;
	}

	template <typename T>
	static inline void copyout(T *uptr, T *kdata, int count)
	{
		// OPT: use alignof for movsw and movsl versions
		ulong bytes = sizeof(*kdata) * count;
		int cause;

		if ((ulong)uptr + bytes >= PHYSMEM_START ||
		    (ulong)uptr + bytes < (ulong)uptr) {
			cause = 1;
			goto bad;
		}
		
		asm("1: rep; movsb\n"
		    "2:\n"
		    ".section .extable,\"a\"\n"
		    ".long 1b\n"
		    ".long 2b\n"
		    ".previous\n" : "+c" (bytes), "+S" (kdata), "+D" (uptr), "=d" (cause));
	
		if (bytes != 0) {
		bad:
			throw_idl(MemoryFault, reinterpret_cast<ulong>(uptr),
			          0, NULL, NULL, cause);
		}
	}

	template <typename T>
	static inline void copyout(T *uptr, T &kdata)
	{
		copyout(uptr, &kdata, sizeof(kdata));
	}
}

#endif
