// io/console/misc.cc -- Generic code relating to kernel debug consoles
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/console.h>
#include <kern/libc.h>
#include <util/spinlock.h>

namespace IO {
namespace Console {
	void Console::write(const char *data, size_t len)
	{
		u64 len64 = len;
		write(Array<octet, ORBMM>((octet *)data, len), &len64);
	}
	
	// All consoles will set this to themselves; the last console
	// to be initialized will be used.  At some point, the console
	// used will be configurable (including the ability to broadcast
	// to multiple consoles).
	
	Console *primary_console;
}
}

static const size_t printf_buffer_size = 4096;
static char printf_buffer[printf_buffer_size];
Lock::SpinLock printf_lock;

// Note: this will not retry with a larger buffer if overflow
// happens.  If you really need that, you'll need to call snprintf
// and primary_console->write yourself.

size_t printf(const char *str, ...)
{
	Lock::AutoSpinLockRecIRQ autolock(printf_lock);

	va_list args;
	va_start(args, str);
	size_t ret = vsnprintf(printf_buffer, printf_buffer_size, str, args);
	va_end(args);
	
	if (ret > printf_buffer_size)
		ret = printf_buffer_size;
	
	IO::Console::primary_console->write(printf_buffer, ret);
	return ret;
}
